from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime, timezone
import itertools
import logging
import numbers
import sys
import uuid
from typing import Callable, Dict, List, overload
from slugify import slugify
from datetime import datetime
from pprint import pprint as pp

logger = logging.getLogger(__name__)


class DataPoint:
    timestamp: datetime = datetime.now()
    Dimensions: dict[str, str] = {}
    Observations: dict[str, float] = {}
    Exogenous: dict[str, float] = {}
    GeographyPoint: GeographyPoint = None

    def __init__(
        self,
        timestamp: datetime,
        observation_name: str,
        observation_value: float,
        observations: dict(str, float) = {},
        dimensions: dict(str, str) = {},
        exogenous: dict(str, float) = {},
        place: GeographyPoint = None,
    ):

        self.Dimensions = {}
        self.Observations = {}
        self.Exogenous = {}
        self.place = None

        self.timestamp = timestamp

        if not observation_name is None and not observation_value is None:
            self = self.add_observation(observation_name, observation_value)

        self.set_place(place)

        for name, value in observations.items():
            self.add_observation(name, value)

        for name, value in dimensions.items():
            self.add_dimension(name, value)

        for name, value in exogenous.items():
            self.add_exogenous(name, value)

    def _add_property_to_collection(
        self, collection: dict, collection_type: str, name: str, value: any
    ) -> DataPoint:

        if not collection_type in ["obs_", "dim_", "exo_", "prop_"]:
            logger.warn(f"Invalid type")
            return self

        if not name or name == "":
            logger.warn(f"Datapoint property was skipped because it has no name.")
            return self

        if name.startswith(collection_type):
            logger.warn(
                f"Datapoint property <{name}> was skipped because it has an invalid name. Starting with <{collection_type}>"
            )
            return self

        if len(name) > 59:
            logger.warn(
                f"Datapoint property <{name}> was skipped because it has an invalid name. <{name}> is too long. Rule : < 59"
            )
            return self

        if value is None:
            logger.warn(
                f"Datapoint property <{name}> was skipped because it has no value."
            )
            return self

        if collection_type in ["obs_", "exo_", "prop_"] and not isinstance(
            value, numbers.Number
        ):
            logger.warn(
                f"Datapoint property <{name}> was skipped because it is not a number."
            )
            return self

        if collection_type == "dim_":
            value = str(value)
            if len(value) > 256:
                logger.warn(
                    f"Datapoint property <{name}> was skipped because its value is too long. Rule : < 256"
                )
                return self

        property_name = f'{slugify(name, separator="-", stopwords=[collection_type])}'
        property_key = collection_type + property_name.lower()

        if property_key == collection_type:
            logger.warn(
                f"Datapoint property <{name}> was skipped because it contains too many special chars."
            )
            return self

        if len(property_key) > 59:
            logger.warn(
                f"Datapoint property <{property_key}> representing ({name}) was skipped because it's name is too long."
            )
            return self

        # if len(property_name)<=2:
        #     logger.info(f"Datapoint property <{name}> was kept but has low semantic value.")

        if property_key in collection.keys():
            logger.warn(
                f"Datapoint property <{name}> with value <{value}> was skipped because it is duplicated."
            )
            return self

        collection[property_key] = value

        return self

    def set_place(self, place: GeographyPoint) -> DataPoint:
        if place is None:
            return self

        self.GeographyPoint = place
        return self

    def add_observation(self, name: str, value: float) -> DataPoint:
        return self._add_property_to_collection(
            collection=self.Observations, collection_type="obs_", name=name, value=value
        )

    def add_exogenous(self, name: str, value: float) -> DataPoint:
        return self._add_property_to_collection(
            collection=self.Exogenous, collection_type="exo_", name=name, value=value
        )

    def add_dimension(self, name: str, value: str) -> DataPoint:
        return self._add_property_to_collection(
            collection=self.Dimensions, collection_type="dim_", name=name, value=value
        )


class GeographyPoint(object):
    latitude: float = 0
    longitude: float = 0
    elevation: int = 0

    def __init__(self, latitude: float, longitude: float, elevation: float = 0):

        if latitude > 90 or latitude < -90:
            raise GeographyPointException("latitude", latitude, "(-90, 90)")

        if longitude > 180 or longitude < -180:
            raise GeographyPointException("longitude", longitude, "(-180, 180)")

        self.latitude = latitude
        self.longitude = longitude
        self.elevation = elevation


class GeographyPointException(Exception):
    def __init__(self, name: str, value: float, range: str):
        self.coodinate_name = name
        self.coodinate_value = value
        self.message = f"The {self.coodinate_name} value {self.coodinate_value} is not valid because {self.coodinate_name} is not in {range} range."
        super().__init__(self.message)

    def __str__(self) -> str:
        return f"{self.coodinate_name} : {self.coodinate_value} -> {self.message}"
